from collections.abc import Mapping
from typing import Mapping as MappingType
from typing import Sequence, Union

from .authorization import AuthorizationContext, Policy, Requirement


class AnonymousRequirement(Requirement):
    """Requires an anonymous user, or service."""

    def handle(self, context: AuthorizationContext):
        identity = context.identity

        if not identity or not identity.is_authenticated():
            context.succeed(self)


class AnonymousPolicy(Policy):
    """Policy that requires an anonymous user, or service."""

    def __init__(self, name: str = "anonymous"):
        super().__init__(name, AnonymousRequirement())


class AuthenticatedRequirement(Requirement):
    """
    Requires an authenticated user, or service. Meaning that an `identity` must be set
    in the authorization context.
    """

    def handle(self, context: AuthorizationContext):
        identity = context.identity

        if identity and identity.is_authenticated():
            context.succeed(self)


RequiredClaimsType = Union[MappingType[str, str], Sequence[str], str]


class ClaimsRequirement(Requirement):
    """Requires an identity with a claims: one or more, optionally with exact values."""

    __slots__ = ("required_claims",)

    def __init__(self, required_claims: RequiredClaimsType):
        if isinstance(required_claims, str):
            required_claims = [required_claims]
        self.required_claims = required_claims

    def handle(self, context: AuthorizationContext):
        identity = context.identity

        if not identity:
            context.fail("Missing identity")
            return

        if isinstance(self.required_claims, Mapping):
            if all(
                identity.has_claim_value(key, value)
                for key, value in self.required_claims.items()
            ):
                context.succeed(self)
        else:
            if all(identity.has_claim(name) for name in self.required_claims):
                context.succeed(self)



from neoteroi.auth.utils import get_running_loop, read_json_data

from . import JWKS, KeysProvider


def _read_openid_configuration(authority: str):
    return read_json_data(authority.rstrip("/") + "/.well-known/openid-configuration")


def read_jwks_from_authority(authority: str) -> JWKS:
    openid_config = _read_openid_configuration(authority)

    if "jwks_uri" not in openid_config:  # pragma: no cover
        raise ValueError("Expected a `jwks_uri` property in the OpenID Configuration")

    jwks = read_json_data(openid_config["jwks_uri"])
    return JWKS.from_dict(jwks)


async def read_jwks_from_authority_async(authority: str) -> JWKS:
    loop = get_running_loop()
    return await loop.run_in_executor(None, lambda: read_jwks_from_authority(authority))


class AuthorityKeysProvider(KeysProvider):
    """
    Kind of KeysProvider that uses the /.well-known/openid-configuration
    discovery endpoint to obtain the `jwks_uri` and the JWKS.
    """

    def __init__(self, authority: str) -> None:
        """
        Creates an instance of AuthorityKeysProvider bound to the given authority.
        """
        super().__init__()
        if not authority:
            raise TypeError("Missing authority")
        self._authority = authority

    @property
    def authority(self) -> str:
        return self._authority

    async def get_keys(self) -> JWKS:
        return await read_jwks_from_authority_async(self._authority)



from abc import ABC
from typing import Any, Iterable, List, Optional, Type, TypeVar, Union

from neoteroi.di import ContainerProtocol

T = TypeVar("T")


class StrategyConfigurationError(Exception):
    """Base class for all configuration errors related to auth strategies."""


class DINotConfiguredError(StrategyConfigurationError):
    def __init__(self) -> None:
        super().__init__(
            "A DI Container is required for this strategy because it needs to activate "
            "types."
        )


class BaseStrategy(ABC):
    def __init__(self, container: Optional[ContainerProtocol] = None) -> None:
        super().__init__()
        self._container = container

    @property
    def container(self) -> ContainerProtocol:
        if self._container is None:
            raise DINotConfiguredError()
        return self._container

    @container.setter
    def container(self, container: ContainerProtocol) -> None:
        self._container = container

    def _get_di_scope(self, scope: Any):
        try:
            return scope._di_scope
        except AttributeError:
            return None

    def _get_instances(self, items: List[Union[T, Type[T]]], scope: Any) -> Iterable[T]:
        """
        Yields instances of types, optionally activated through dependency injection.

        If the given context has a DI scope defined in "_di_scope" attribute, it is
        used to support scoped services.
        """
        scope = self._get_di_scope(scope)
        for obj in items:
            if isinstance(obj, type):
                # a container is required
                yield self.container.resolve(obj, scope=scope)
            else:
                yield obj



import asyncio
import json
import logging
import sys
import urllib.error
import urllib.request


def read_json_data(url: str):
    with urllib.request.urlopen(url) as response:
        return json.loads(response.read())


def get_logger():
    return logging.getLogger("auth-jwts")


def get_running_loop():  # pragma: no cover
    if sys.version_info[:2] <= (3, 7):
        # For Python 3.6
        return asyncio._get_running_loop()
    else:
        return asyncio.get_running_loop()



import inspect
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Any, List, Optional, Sequence, Type, Union

from neoteroi.di import ContainerProtocol

from neoteroi.auth.abc import BaseStrategy


class Identity:
    """
    Represents the characteristics of a person or a thing in the context of an
    application. It can be a user interacting with an app, or a technical account.
    """

    def __init__(
        self,
        claims: Optional[dict] = None,
        authentication_mode: Optional[str] = None,
    ):
        self.claims = claims or {}
        self.authentication_mode = authentication_mode
        self.access_token: Optional[str] = None
        self.refresh_token: Optional[str] = None

    @property
    def sub(self) -> Optional[str]:
        return self["sub"]

    def is_authenticated(self) -> bool:
        return bool(self.authentication_mode)

    def __getitem__(self, item):
        return self.claims[item]

    def has_claim(self, name: str) -> bool:
        return name in self.claims

    def has_claim_value(self, name: str, value: str) -> bool:
        return self.claims.get(name) == value


class User(Identity):
    @property
    def id(self) -> Optional[str]:
        return self["id"] or self.sub

    @property
    def name(self) -> Optional[str]:
        return self["name"]

    @property
    def email(self) -> Optional[str]:
        return self["email"]


class AuthenticationHandler(ABC):
    """Base class for types that implement authentication logic."""

    @property
    def scheme(self) -> str:
        """Returns the name of the Authentication Scheme used by this handler."""
        return self.__class__.__name__

    @abstractmethod
    def authenticate(self, context: Any) -> Optional[Identity]:
        """Obtains an identity from a context."""


@lru_cache(maxsize=None)
def _is_async_handler(handler_type: Type[AuthenticationHandler]) -> bool:
    # Faster alternative to using inspect.iscoroutinefunction without caching
    # Note: this must be used on Types - not instances!
    return inspect.iscoroutinefunction(handler_type.authenticate)


AuthenticationHandlerConfType = Union[
    AuthenticationHandler, Type[AuthenticationHandler]
]


class AuthenticationSchemesNotFound(ValueError):
    def __init__(
        self, configured_schemes: Sequence[str], required_schemes: Sequence[str]
    ):
        super().__init__(
            "Could not find authentication handlers for required schemes: "
            f'{", ".join(required_schemes)}. '
            f'Configured schemes are: {", ".join(configured_schemes)}'
        )


class AuthenticationStrategy(BaseStrategy):
    def __init__(
        self,
        *handlers: AuthenticationHandlerConfType,
        container: Optional[ContainerProtocol] = None,
    ):
        super().__init__(container)
        self.handlers = list(handlers)

    def add(self, handler: AuthenticationHandlerConfType) -> "AuthenticationStrategy":
        self.handlers.append(handler)
        return self

    def __iadd__(
        self, handler: AuthenticationHandlerConfType
    ) -> "AuthenticationStrategy":
        self.handlers.append(handler)
        return self

    def _get_handlers_by_schemes(
        self,
        authentication_schemes: Optional[Sequence[str]] = None,
        context: Any = None,
    ) -> List[AuthenticationHandler]:
        if not authentication_schemes:
            return list(self._get_instances(self.handlers, context))

        handlers = [
            handler
            for handler in self._get_instances(self.handlers, context)
            if handler.scheme in authentication_schemes
        ]

        if not handlers:
            raise AuthenticationSchemesNotFound(
                [
                    handler.scheme
                    for handler in self._get_instances(self.handlers, context)
                ],
                authentication_schemes,
            )

        return handlers

    async def authenticate(
        self, context: Any, authentication_schemes: Optional[Sequence[str]] = None
    ) -> Optional[Identity]:
        """
        Tries to obtain the user for a context, applying authentication rules.
        """
        if not context:
            raise ValueError("Missing context to evaluate authentication")

        for handler in self._get_handlers_by_schemes(authentication_schemes, context):
            if _is_async_handler(type(handler)):
                identity = await handler.authenticate(context)  # type: ignore
            else:
                identity = handler.authenticate(context)

            if identity:
                try:
                    context.identity = identity
                except AttributeError:
                    pass
                return identity
        return None



import time
from typing import Optional

from . import JWKS, KeysProvider


class CachingKeysProvider(KeysProvider):
    """
    Kind of KeysProvider that can cache the result of another KeysProvider.
    """

    def __init__(self, keys_provider: KeysProvider, cache_time: float) -> None:
        """
        Creates a new instance of CachingKeysProvider bound to a given KeysProvider,
        and caching its result up to an optional amount of seconds described by
        cache_time. Expiration is disabled if `cache_time` <= 0.
        """
        super().__init__()

        if not keys_provider:
            raise TypeError("Missing KeysProvider")

        self._keys: Optional[JWKS] = None
        self._cache_time = cache_time
        self._last_fetch_time: float = 0
        self._keys_provider = keys_provider

    @property
    def keys_provider(self) -> KeysProvider:
        return self._keys_provider

    async def _fetch_keys(self) -> JWKS:
        self._keys = await self._keys_provider.get_keys()
        self._last_fetch_time = time.time()
        return self._keys

    async def get_keys(self) -> JWKS:
        if self._keys is not None:
            if self._cache_time > 0 and (
                time.time() - self._last_fetch_time >= self._cache_time
            ):
                pass
            else:
                return self._keys
        return await self._fetch_keys()



from typing import Any, Dict, Optional, Sequence

import jwt
from jwt.exceptions import InvalidIssuerError, InvalidTokenError

from ..jwks import JWK, JWKS, KeysProvider
from ..jwks.caching import CachingKeysProvider
from ..jwks.openid import AuthorityKeysProvider
from ..jwks.urls import URLKeysProvider
from ..utils import get_logger


class OAuthException(Exception):
    """Base class for exception risen when there is an issue related to OAuth."""


class InvalidAccessToken(Exception):
    def __init__(self, details=""):
        if details:
            message = "Invalid access token: " + details
        else:
            message = "Invalid access token."
        super().__init__(message)


def get_kid(token: str) -> Optional[str]:
    """
    Extracts a kid (key id) from a JWT.
    """
    headers = jwt.get_unverified_header(token)
    if not headers:  # pragma: no cover
        raise InvalidAccessToken("missing headers")
    return headers.get("kid")


class JWTValidator:
    def __init__(
        self,
        *,
        valid_issuers: Sequence[str],
        valid_audiences: Sequence[str],
        authority: Optional[str] = None,
        algorithms: Sequence[str] = ["RS256"],
        require_kid: bool = True,
        keys_provider: Optional[KeysProvider] = None,
        keys_url: Optional[str] = None,
        cache_time: float = 10800
    ) -> None:
        """
        Creates a new instance of JWTValidator. This class only supports validating
        access tokens signed using asymmetric keys and handling JWKs of RSA type.

        Parameters
        ----------
        valid_issuers : Sequence[str]
            Sequence of acceptable issuers (iss).
        valid_audiences : Sequence[str]
            Sequence of acceptable audiences (aud).
        authority : Optional[str], optional
            If provided, keys are obtained from a standard well-known endpoint.
            This parameter is ignored if `keys_provider` is given.
        algorithms : Sequence[str], optional
            Sequence of acceptable algorithms, by default ["RS256"].
        require_kid : bool, optional
            According to the specification, a key id is optional in JWK. However,
            this parameter lets control whether access tokens missing `kid` in their
            headers should be handled or rejected. By default True, thus only JWTs
            having `kid` header are accepted.
        keys_provider : Optional[KeysProvider], optional
            If provided, the exact `KeysProvider` to be used when fetching keys.
            By default None
        keys_url : Optional[str], optional
            If provided, keys are obtained from the given URL through HTTP GET.
            This parameter is ignored if `keys_provider` is given.
        cache_time : float, optional
            If >= 0, JWKS are cached in memory and stored for the given amount in
            seconds. By default 10800 (3 hours).
        """
        if keys_provider:
            pass
        elif authority:
            keys_provider = AuthorityKeysProvider(authority)
        elif keys_url:
            keys_provider = URLKeysProvider(keys_url)

        if keys_provider is None:
            raise TypeError(
                "Missing `keys_provider`, either provide a `url` source, "
                "`authority`, or `keys_provider`."
            )

        if cache_time:
            keys_provider = CachingKeysProvider(keys_provider, cache_time)

        self._valid_issuers = list(valid_issuers)
        self._valid_audiences = list(valid_audiences)
        self._algorithms = list(algorithms)
        self._keys_provider: KeysProvider = keys_provider
        self.require_kid = require_kid
        self.logger = get_logger()

    async def get_jwks(self) -> JWKS:
        return await self._keys_provider.get_keys()

    async def get_jwk(self, kid: str) -> JWK:
        jwks = await self.get_jwks()

        for jwk in jwks.keys:
            if jwk.kid is not None and jwk.kid == kid:
                return jwk
        raise InvalidAccessToken("kid not recognized")

    def _validate_jwt_by_key(
        self, access_token: str, jwk: JWK
    ) -> Optional[Dict[str, Any]]:
        for issuer in self._valid_issuers:
            try:
                return jwt.decode(
                    access_token,
                    jwk.pem,  # type: ignore
                    verify=True,
                    algorithms=self._algorithms,
                    audience=self._valid_audiences,
                    issuer=issuer,
                )
            except InvalidIssuerError:
                # pass, because the application might support more than one issuer;
                # note that token verification might fail for several other reasons
                # that are not catched (e.g. expired signature)
                pass
            except InvalidTokenError as exc:
                self.logger.debug("Invalid access token: ", exc_info=exc)
                return None
        return None

    async def validate_jwt(self, access_token: str) -> Dict[str, Any]:
        """
        Validates the given JWT and returns its payload. This method throws exception
        if the JWT is not valid (i.e. its signature cannot be verified, for example
        because the JWT expired).
        """
        kid = get_kid(access_token)
        if kid is None and self.require_kid:
            # A key id is optional according to the specification,
            # but here we expect a kid by default.
            # Disabling require_kid makes this method less efficient.
            raise InvalidAccessToken("Missing key id (kid).")

        if kid is None:
            # Unoptimal scenario: the identity provider does not handle key ids,
            # thus if more than one JWK is configured in the JWKS, we need to cycle
            # and attempt each of them
            jwks = await self.get_jwks()

            for jwk in jwks.keys:
                data = self._validate_jwt_by_key(access_token, jwk)
                if data is not None:
                    return data
        else:
            # Preferred scenario: the identity provider handles key ids,
            # thus we can validate an access token using an exact key
            jwk = await self.get_jwk(kid)
            data = self._validate_jwt_by_key(access_token, jwk)
            if data is not None:
                return data

        raise InvalidAccessToken()



import inspect
from abc import ABC, abstractmethod
from functools import lru_cache, wraps
from typing import Any, Callable, Iterable, List, Optional, Sequence, Set, Type, Union

from neoteroi.di import ContainerProtocol

from neoteroi.auth.abc import BaseStrategy
from neoteroi.auth.authentication import Identity


class AuthorizationError(Exception):
    pass


class AuthorizationConfigurationError(Exception):
    pass


class PolicyNotFoundError(AuthorizationConfigurationError, RuntimeError):
    def __init__(self, name: str):
        super().__init__(f"Cannot find policy with name {name}")


class Requirement(ABC):
    """Base class for authorization requirements."""

    def __str__(self):
        return self.__class__.__name__

    @abstractmethod
    async def handle(self, context: "AuthorizationContext"):
        """Handles this requirement for a given context."""


RequirementConfType = Union[Requirement, Type[Requirement]]


@lru_cache(maxsize=None)
def _is_async_handler(handler_type: Type[Requirement]) -> bool:
    # Faster alternative to using inspect.iscoroutinefunction without caching
    # Note: this must be used on Types - not instances!
    return inspect.iscoroutinefunction(handler_type.handle)


class UnauthorizedError(AuthorizationError):
    def __init__(
        self,
        forced_failure: Optional[str],
        failed_requirements: Sequence[Requirement],
        scheme: Optional[str] = None,
        error: Optional[str] = None,
        error_description: Optional[str] = None,
    ):
        """
        Creates a new instance of UnauthorizedError, with details.

        :param forced_failure: if applicable, the reason for a forced failure.
        :param failed_requirements: a sequence of requirements that failed.
        :param scheme: optional authentication scheme that should be used.
        :param error: optional error short text.
        :param error_description: optional error details.
        """
        super().__init__(self._get_message(forced_failure, failed_requirements))
        self.failed = forced_failure
        self.failed_requirements = failed_requirements
        self.scheme = scheme
        self.error = error
        self.error_description = error_description

    @staticmethod
    def _get_message(forced_failure, failed_requirements):
        if forced_failure:
            return (
                "The user is not authorized to perform the selected action."
                + f" {forced_failure}."
            )

        if failed_requirements:
            errors = ", ".join(str(requirement) for requirement in failed_requirements)
            return (
                f"The user is not authorized to perform the selected action. "
                f"Failed requirements: {errors}."
            )
        return "Unauthorized"


class AuthorizationContext:

    __slots__ = ("identity", "requirements", "_succeeded", "_failed_forced")

    def __init__(self, identity: Identity, requirements: Sequence[Requirement]):
        self.identity = identity
        self.requirements = requirements
        self._succeeded: Set[Requirement] = set()
        self._failed_forced: Optional[str] = None

    @property
    def pending_requirements(self) -> List[Requirement]:
        return [item for item in self.requirements if item not in self._succeeded]

    @property
    def has_succeeded(self) -> bool:
        if self._failed_forced:
            return False
        return all(requirement in self._succeeded for requirement in self.requirements)

    @property
    def forced_failure(self) -> Optional[str]:
        return None if self._failed_forced is None else str(self._failed_forced)

    def fail(self, reason: str):
        """
        Called to indicate that this authorization context has failed.
        Forces failure, regardless of succeeded requirements.
        """
        self._failed_forced = reason or "Authorization failed."

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.clear()

    def succeed(self, requirement: Requirement):
        """Marks the given requirement as succeeded for this authorization context."""
        self._succeeded.add(requirement)

    def clear(self):
        self._failed_forced = None
        self._succeeded.clear()


class Policy:
    """
    Represents an authorization policy, with a set of authorization rules.
    """

    __slots__ = ("name", "requirements")

    def __init__(self, name: str, *requirements: RequirementConfType):
        self.name = name
        self.requirements = list(requirements) or []

    def _valid_requirement(self, obj):
        if not isinstance(obj, Requirement) or (
            isinstance(obj, type) and not issubclass(obj, Requirement)
        ):
            raise ValueError(
                "Only instances, or types, of Requirement can be added to the policy."
            )

    def add(self, requirement: RequirementConfType) -> "Policy":
        self._valid_requirement(requirement)
        self.requirements.append(requirement)
        return self

    def __iadd__(self, other: RequirementConfType):
        self._valid_requirement(other)
        self.requirements.append(other)
        return self

    def __repr__(self):
        return f'<Policy "{self.name}" at {id(self)}>'


class AuthorizationStrategy(BaseStrategy):
    def __init__(
        self,
        *policies: Policy,
        container: Optional[ContainerProtocol] = None,
        default_policy: Optional[Policy] = None,
        identity_getter: Optional[Callable[..., Identity]] = None,
    ):
        super().__init__(container)
        self.policies = list(policies)
        self.default_policy = default_policy
        self.identity_getter = identity_getter

    def get_policy(self, name: str) -> Optional[Policy]:
        for policy in self.policies:
            if policy.name == name:
                return policy
        return None

    def add(self, policy: Policy) -> "AuthorizationStrategy":
        self.policies.append(policy)
        return self

    def __iadd__(self, policy: Policy) -> "AuthorizationStrategy":
        self.policies.append(policy)
        return self

    def with_default_policy(self, policy: Policy) -> "AuthorizationStrategy":
        self.default_policy = policy
        return self

    async def authorize(
        self, policy_name: Optional[str], identity: Identity, scope: Any = None
    ):
        if policy_name:
            policy = self.get_policy(policy_name)

            if not policy:
                raise PolicyNotFoundError(policy_name)

            await self._handle_with_policy(policy, identity, scope)
        else:
            if self.default_policy:
                await self._handle_with_policy(self.default_policy, identity, scope)
                return

            if not identity:
                raise UnauthorizedError("Missing identity", [])
            if not identity.is_authenticated():
                raise UnauthorizedError("The resource requires authentication", [])

    def _get_requirements(self, policy: Policy, scope: Any) -> Iterable[Requirement]:
        yield from self._get_instances(policy.requirements, scope)

    async def _handle_with_policy(self, policy: Policy, identity: Identity, scope: Any):
        with AuthorizationContext(
            identity, list(self._get_requirements(policy, scope))
        ) as context:

            for requirement in context.requirements:
                if _is_async_handler(type(requirement)):  # type: ignore
                    await requirement.handle(context)
                else:
                    requirement.handle(context)  # type: ignore

            if not context.has_succeeded:
                raise UnauthorizedError(
                    context.forced_failure, context.pending_requirements
                )

    async def _handle_with_identity_getter(
        self, policy_name: Optional[str], *args, **kwargs
    ):
        if self.identity_getter is None:
            raise TypeError("Missing identity getter function.")
        await self.authorize(policy_name, self.identity_getter(*args, **kwargs))

    def __call__(self, policy: Optional[str] = None):
        """
        Decorates a function to apply authorization logic on each call.
        """

        def decorator(fn):
            @wraps(fn)
            async def wrapper(*args, **kwargs):
                await self._handle_with_identity_getter(policy, *args, **kwargs)
                return await fn(*args, **kwargs)

            return wrapper

        return decorator



import base64
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers

from neoteroi.auth.errors import UnsupportedFeatureError


def _raise_if_missing(value: dict, *keys: str) -> None:
    for key in keys:
        if key not in value or not bool(value[key]):
            raise ValueError(f"Missing {key}")


class KeyType(Enum):
    EC = "EC"
    RSA = "RSA"
    OCT = "oct"
    OKP = "OKP"

    @classmethod
    def from_str(cls, value: str) -> "KeyType":
        if not value:
            raise ValueError("Missing key type (kty)")
        try:
            return cls[value.upper()]
        except KeyError:
            raise ValueError(f"Invalid JWT kty parameter: {value}")


@dataclass
class JWK:
    """
    This class provides an interface to a JSON Web Key.
    A JSON Web Key (JWK) is a JavaScript Object Notation (JSON) data
    structure that represents a cryptographic key.

    For more information: https://datatracker.ietf.org/doc/html/rfc7517
    """

    kty: KeyType
    n: str
    e: str
    pem: bytes
    kid: Optional[str] = None

    @classmethod
    def from_dict(cls, value) -> "JWK":
        key_type = KeyType.from_str(value.get("kty"))

        if key_type != KeyType.RSA:
            raise UnsupportedFeatureError("This library supports only RSA public keys.")

        _raise_if_missing(value, "n", "e")
        return cls(
            kty=key_type,
            n=value["n"],
            e=value["e"],
            kid=value.get("kid"),
            pem=rsa_pem_from_n_and_e(value["n"], value["e"]),
        )


@dataclass
class JWKS:
    keys: List[JWK]

    @classmethod
    def from_dict(cls, value) -> "JWKS":
        if "keys" not in value:
            raise ValueError("Invalid JWKS structure, missing `keys` property.")
        return cls(keys=[JWK.from_dict(item) for item in value["keys"]])


class KeysProvider(ABC):
    """
    Base for classes that can provide a JWKS from a source.
    """

    @abstractmethod
    async def get_keys(self) -> JWKS:
        """Returns a JWKS."""


class InMemoryKeysProvider(KeysProvider):
    """
    Type of keys provider that stores keys in memory.
    """

    def __init__(self, keys: JWKS) -> None:
        """
        Creates a new instance of InMemoryKeysProvider bound to the given JWKS.

        Parameters
        ----------
        keys : JWKS
            Exact keys handled by this instance.
        """
        super().__init__()
        self._keys = keys

    async def get_keys(self) -> JWKS:
        return self._keys


def _ensure_bytes(key):
    if isinstance(key, str):
        key = key.encode("utf-8")
    return key


def _decode_value(val):
    decoded = base64.urlsafe_b64decode(_ensure_bytes(val) + b"==")
    return int.from_bytes(decoded, "big")


def rsa_pem_from_n_and_e(n: str, e: str) -> bytes:
    return (
        RSAPublicNumbers(n=_decode_value(n), e=_decode_value(e))
        .public_key(default_backend())
        .public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
    )



from .authentication import (
    AuthenticationHandler,
    AuthenticationHandlerConfType,
    AuthenticationSchemesNotFound,
    AuthenticationStrategy,
    Identity,
    User,
)
from .authorization import (
    AuthorizationConfigurationError,
    AuthorizationContext,
    AuthorizationError,
    AuthorizationStrategy,
    Policy,
    PolicyNotFoundError,
    Requirement,
    RequirementConfType,
    UnauthorizedError,
)

__all__ = [
    "AuthenticationHandlerConfType",
    "AuthenticationSchemesNotFound",
    "AuthorizationConfigurationError",
    "AuthenticationStrategy",
    "AuthorizationStrategy",
    "AuthenticationHandler",
    "AuthorizationError",
    "Identity",
    "User",
    "Policy",
    "PolicyNotFoundError",
    "Requirement",
    "RequirementConfType",
    "UnauthorizedError",
    "AuthorizationContext",
]



from neoteroi.auth.utils import get_running_loop, read_json_data

from . import JWKS, KeysProvider


def read_jwks_from_url(url: str) -> JWKS:
    jwks = read_json_data(url)
    return JWKS.from_dict(jwks)


async def read_jwks_from_url_async(url: str) -> JWKS:
    loop = get_running_loop()
    return await loop.run_in_executor(None, lambda: read_jwks_from_url(url))


class URLKeysProvider(KeysProvider):
    """
    Kind of KeysProvider that obtains JWKS from a given URL.
    """

    def __init__(self, url: str) -> None:
        """
        Creates an instance of URLKeysProvider bound to the given URL.
        """
        super().__init__()
        if not url:
            raise TypeError("Missing URL")
        self._url = url

    @property
    def url(self) -> str:
        return self._url

    async def get_keys(self) -> JWKS:
        return await read_jwks_from_url_async(self._url)
